from __future__ import annotations

import datetime
import json
import os
import posixpath
import uuid
from dataclasses import dataclass

from fal.toolkit.file.types import FileData, FileRepository

DEFAULT_URL_TIMEOUT = 60 * 15  # 15 minutes


@dataclass
class GoogleStorageRepository(FileRepository):
    bucket_name: str = "fal_file_storage"
    url_expiration: int | None = DEFAULT_URL_TIMEOUT
    gcp_account_json: str | None = None
    folder: str = ""

    _storage_client = None
    _bucket = None

    def __post_init__(self):
        gcp_account_json = self.gcp_account_json
        if gcp_account_json is None:
            gcp_account_json = os.environ.get("GCLOUD_SA_JSON")
            if gcp_account_json is None:
                raise Exception("GCLOUD_SA_JSON environment secret is not set")

        gcp_account_info = json.loads(gcp_account_json)

        from google.cloud.storage import Client

        self._storage_client = Client.from_service_account_info(gcp_account_info)
        self._bucket = self._storage_client.get_bucket(self.bucket_name)

    @property
    def storage_client(self):
        if self._storage_client is None:
            raise Exception("Google Storage client is not initialized")

        return self._storage_client

    @property
    def bucket(self):
        if self._bucket is None:
            raise Exception("Google Storage bucket is not initialized")

        return self._bucket

    def save(self, data: FileData) -> str:
        destination_path = posixpath.join(
            self.folder,
            f"{uuid.uuid4().hex}_{data.file_name}",
        )

        gcp_blob = self.bucket.blob(destination_path)
        gcp_blob.upload_from_string(data.data, content_type=data.content_type)

        if self.url_expiration is None:
            return gcp_blob.public_url

        return gcp_blob.generate_signed_url(
            version="v4",
            expiration=datetime.timedelta(minutes=self.url_expiration),
            method="GET",
        )



from unittest.mock import MagicMock, patch

import pytest
from fal.cli.main import parse_args
from fal.cli.run import _run


def test_run():
    args = parse_args(["run", "/my/path.py::myfunc"])
    assert args.func == _run
    assert args.func_ref == ("/my/path.py", "myfunc")


@pytest.fixture
def mock_parse_pyproject_toml():
    return {
        "apps": {
            "my-app": {
                "ref": "src/my_app/inference.py::MyApp",
                "auth": "shared",
            },
            # auth is not provided for another-app
            "another-app": {
                "ref": "src/another_app/inference.py::AnotherApp",
            },
        }
    }


def mocked_fal_serverless_host(host):
    mock = MagicMock()
    mock.host = host
    return mock


def mock_args(host, func_ref: tuple[str]):
    args = MagicMock()
    args.host = host
    args.func_ref = func_ref
    return args


@patch("fal.cli._utils.find_pyproject_toml", return_value="pyproject.toml")
@patch("fal.cli._utils.parse_pyproject_toml")
@patch("fal.api.FalServerlessHost")
@patch("fal.utils.load_function_from")
def test_run_with_toml_success(
    mock_load_function_from,
    mock_fal_serverless_host,
    mock_parse_toml,
    mock_find_toml,
    mock_parse_pyproject_toml,
):
    mock_parse_toml.return_value = mock_parse_pyproject_toml

    host = mocked_fal_serverless_host("my-host")
    mock_fal_serverless_host.return_value = host

    args = mock_args(func_ref=("my-app", None), host=host)

    _run(args)

    # Ensure the correct app is deployed
    mock_load_function_from.assert_called_once_with(
        host, "src/my_app/inference.py", "MyApp"
    )


@patch("fal.cli._utils.find_pyproject_toml", return_value="pyproject.toml")
@patch("fal.cli._utils.parse_pyproject_toml")
@patch("fal.utils.load_function_from")
def test_run_with_toml_app_not_found(
    mock_load_function_from, mock_parse_toml, mock_find_toml, mock_parse_pyproject_toml
):
    mock_parse_toml.return_value = mock_parse_pyproject_toml
    args = mock_args(func_ref=("non-existent-app", None), host="my-host")

    with pytest.raises(
        ValueError, match="App non-existent-app not found in pyproject.toml"
    ):
        _run(args)

    # Ensure _run_from_reference was not called
    mock_load_function_from.assert_not_called()


@patch("fal.cli._utils.find_pyproject_toml", return_value="pyproject.toml")
@patch("fal.cli._utils.parse_pyproject_toml")
@patch("fal.utils.load_function_from")
def test_run_with_toml_missing_ref_key(
    mock_load_function_from, mock_parse_toml, mock_find_toml
):
    mock_parse_toml.return_value = {
        "apps": {
            "my-app": {
                "auth": "shared",
            }
        }
    }

    args = mock_args(func_ref=("my-app", None), host="my-host")

    with pytest.raises(
        ValueError, match="App my-app does not have a ref key in pyproject.toml"
    ):
        _run(args)

    # Ensure _run_from_reference was not called
    mock_load_function_from.assert_not_called()



from __future__ import annotations

import sys
from datetime import datetime, timezone

from isolate.logs import Log, LogLevel, LogSource
from structlog.dev import ConsoleRenderer
from structlog.typing import EventDict

from .style import LEVEL_STYLES

_renderer = ConsoleRenderer(level_styles=LEVEL_STYLES)


class IsolateLogPrinter:
    debug: bool

    def __init__(self, debug: bool = False) -> None:
        self.debug = debug
        self._current_source: LogSource | None = None

    def _maybe_print_header(self, source: LogSource):
        from fal.console import console

        if source == self._current_source:
            return

        msg = {
            LogSource.BUILDER: "Building the environment",
            LogSource.BRIDGE: "Unpacking user code",
            LogSource.USER: "Running",
        }.get(source)

        if msg:
            console.print(f"==> {msg}", style="bold green")

        self._current_source = source

    def print(self, log: Log):
        if log.level < LogLevel.INFO and not self.debug:
            return

        self._maybe_print_header(log.source)

        if log.source == LogSource.USER:
            stream = sys.stderr if log.level == LogLevel.STDERR else sys.stdout
            print(log.message, file=stream)
            return

        level = str(log.level)

        if hasattr(log, "timestamp"):
            timestamp = log.timestamp
        else:
            # Default value for timestamp if user has old `isolate` version.
            # Even if the controller version is controller by us, which means that
            # the timestamp is being sent in the gRPC message.
            # The `isolate` version users interpret that message with is out of our
            # control. So we need to handle this case.
            timestamp = datetime.now(timezone.utc)

        event: EventDict = {
            "event": log.message,
            "level": level,
            "timestamp": timestamp.strftime("%Y-%m-%d %H:%M:%S.%f")[:-3],
        }
        if self.debug and log.bound_env and log.bound_env.key != "global":
            event["bound_env"] = log.bound_env.key

        # Use structlog processors to get consistent output with local logs
        message = _renderer.__call__(logger={}, name=level, event_dict=event)
        print(message)



from typing import Any

import fal

from . import filter_by
from .nsfw_filter.model import get_model


@fal.cached
def load_safety_checker():
    import torch
    from diffusers.pipelines.stable_diffusion.safety_checker import (
        StableDiffusionSafetyChecker,
    )
    from transformers import AutoFeatureExtractor

    feature_extractor = AutoFeatureExtractor.from_pretrained(
        "CompVis/stable-diffusion-safety-checker",
        torch_dtype="float16",
    )
    safety_checker = StableDiffusionSafetyChecker.from_pretrained(
        "CompVis/stable-diffusion-safety-checker",
        torch_dtype=torch.float16,
    ).to("cuda")

    return feature_extractor, safety_checker


def run_safety_checker(
    pil_images: list[object],
) -> list[bool]:
    import numpy as np
    import torch

    feature_extractor, safety_checker = load_safety_checker()

    safety_checker_input = feature_extractor(pil_images, return_tensors="pt").to("cuda")

    np_image = [np.array(val) for val in pil_images]

    _, has_nsfw_concept = safety_checker(
        images=np_image,
        clip_input=safety_checker_input.pixel_values.to(torch.float16),
    )

    return has_nsfw_concept


def run_safety_checker_v2(pil_images: list, nsfw_threshold: float = 0.5) -> list[bool]:
    import torch

    model, processor = get_model()

    has_nsfw_concept = []

    with torch.no_grad():
        for pil_image in pil_images:
            inputs = processor(
                images=pil_image.convert("RGB"),
                return_tensors="pt",
            )
            outputs = model(**inputs)
            logits = (
                outputs.logits.squeeze()
            )  # Remove batch dimension to simplify indexing

            # Apply softmax to convert logits to probabilities
            probabilities = torch.softmax(logits, dim=0)

            nsfw_class_index = model.config.label2id.get(
                "nsfw", None
            )  # Replace "NSFW" with the exact class name if different

            # Validate that NSFW class index is found
            if nsfw_class_index is not None:
                nsfw_probability = probabilities[int(nsfw_class_index)].item()
                print("NSFW probability:", nsfw_probability)
                has_nsfw_concept.append(nsfw_probability > nsfw_threshold)
            else:
                raise ValueError("NSFW class not found in model output.")

    return has_nsfw_concept


def postprocess_images(
    pil_images: list[object],
    enable_safety_checker: bool = True,
    safety_checker_version: int = 2,
) -> dict[str, Any]:
    outputs: dict[str, list[Any]] = {
        "images": pil_images,
    }

    if enable_safety_checker:
        safety_checker_fn = (
            run_safety_checker_v2 if safety_checker_version == 2 else run_safety_checker
        )
        outputs["has_nsfw_concepts"] = safety_checker_fn(pil_images)  # type: ignore
    else:
        outputs["has_nsfw_concepts"] = [False] * len(pil_images)

    outputs["images"] = filter_by(
        outputs["has_nsfw_concepts"],
        outputs["images"],
    )

    return outputs



from __future__ import annotations

from functools import partial

import pytest
from fal import function


@pytest.fixture(scope="function")
def isolated_client():
    return partial(function, machine_type="XS", keep_alive=0)



import argparse
import sys

import rich_argparse


class FalParserExit(Exception):
    def __init__(self, status=0):
        self.status = status


class RefAction(argparse.Action):
    def __init__(self, *args, **kwargs):
        kwargs.setdefault("default", (None, None))
        super().__init__(*args, **kwargs)

    @classmethod
    def split_ref(cls, value):
        if isinstance(value, tuple):
            return value

        if value.find("::") > 1:
            return value.split("::", 1)

        return value, None

    def __call__(self, parser, args, values, option_string=None):  # noqa: ARG002
        file_path, obj_path = self.split_ref(values)
        setattr(args, self.dest, (file_path, obj_path))


class DictAction(argparse.Action):
    def __init__(self, *args, **kwargs):
        kwargs.setdefault("metavar", "<name>=<value>")
        super().__init__(*args, **kwargs)

    def __call__(self, parser, args, values, option_string=None):  # noqa: ARG002
        d = getattr(args, self.dest) or {}

        if isinstance(values, list):
            kvs = values
        else:
            kvs = [values]

        for kv in kvs:
            parts = kv.split("=", 1)
            if len(parts) != 2:
                raise argparse.ArgumentError(
                    self,
                    f'Could not parse argument "{values}" as k1=v1 k2=v2 ... format',
                )
            key, value = parts
            d[key] = value

        setattr(args, self.dest, d)


def _find_parser(parser, func):
    defaults = parser._defaults
    if not func or func == defaults.get("func"):
        return parser

    actions = parser._actions
    for action in actions:
        if not isinstance(action.choices, dict):
            continue
        for subparser in action.choices.values():
            par = _find_parser(subparser, func)
            if par:
                return par
    return None


class FalParser(argparse.ArgumentParser):
    def __init__(self, *args, **kwargs):
        kwargs.setdefault("formatter_class", rich_argparse.RawTextRichHelpFormatter)
        super().__init__(*args, **kwargs)

    def exit(self, status=0, message=None):
        if message:
            self._print_message(message, sys.stderr)
        raise FalParserExit(status)

    def parse_args(self, args=None, namespace=None):
        args, argv = self.parse_known_args(args, namespace)
        if argv:
            parser = _find_parser(self, getattr(args, "func", None)) or self
            parser.error("unrecognized arguments: %s" % " ".join(argv))
        return args


class FalClientParser(FalParser):
    def __init__(self, *args, **kwargs):
        from fal.flags import GRPC_HOST

        super().__init__(*args, **kwargs)
        self.add_argument(
            "--host",
            default=GRPC_HOST,
            help=argparse.SUPPRESS,
        )



import asyncio
import json
import secrets
import subprocess
import time
from contextlib import contextmanager, suppress
from datetime import datetime
from typing import Generator

import fal
import fal.api as api
import httpx
import pytest
from fal import apps
from fal.app import AppClient
from fal.cli.deploy import _get_user
from fal.container import ContainerImage
from fal.exceptions import AppException, FieldException, RequestCancelledException
from fal.exceptions._cuda import _CUDA_OOM_MESSAGE, _CUDA_OOM_STATUS_CODE
from fal.rest_client import REST_CLIENT
from fal.workflows import Workflow
from fastapi import WebSocket
from httpx import HTTPStatusError
from isolate.backends.common import active_python
from openapi_fal_rest.api.applications import app_metadata
from pydantic import BaseModel
from pydantic import __version__ as pydantic_version


class Input(BaseModel):
    lhs: int
    rhs: int
    wait_time: int = 0


class StatefulInput(BaseModel):
    value: int


class Output(BaseModel):
    result: int


actual_python = active_python()


def git_revision_short_hash() -> str:
    return (
        subprocess.check_output(["git", "rev-parse", "--short", "HEAD"])
        .decode("ascii")
        .strip()
    )


@fal.function(
    keep_alive=60,
    machine_type="S",
    serve=True,
    max_concurrency=1,
    requirements=[f"pydantic=={pydantic_version}"],
)
def addition_app(input: Input) -> Output:
    print("starting...")
    for _ in range(input.wait_time):
        print("sleeping...")
        time.sleep(1)

    return Output(result=input.lhs + input.rhs)


nomad_addition_app = addition_app.on(_scheduler="nomad")


@fal.function(
    kind="container",
    image=ContainerImage.from_dockerfile_str(
        f"FROM python:{actual_python}-slim\n# {git_revision_short_hash()}",
    ),
    keep_alive=60,
    machine_type="S",
    serve=True,
    max_concurrency=1,
)
def container_addition_app(input: Input) -> Output:
    print("starting...")
    for _ in range(input.wait_time):
        print("sleeping...")
        time.sleep(1)

    return Output(result=input.lhs + input.rhs)


@fal.function(
    keep_alive=300,
    requirements=["fastapi", "uvicorn", "pydantic==1.10.12"],
    machine_type="S",
    max_concurrency=1,
    max_multiplexing=30,
    exposed_port=8000,
)
def calculator_app():
    from fastapi import FastAPI
    from fastapi.middleware.cors import CORSMiddleware
    from uvicorn import run

    app = FastAPI()

    def _wait(wait_time: int):
        print("starting...")
        for _ in range(wait_time):
            print("sleeping...")
            time.sleep(1)

    @app.post("/add")
    def add(input: Input) -> Output:
        _wait(input.wait_time)
        return Output(result=input.lhs + input.rhs)

    @app.post("/subtract")
    def subtract(input: Input) -> Output:
        _wait(input.wait_time)
        return Output(result=input.lhs - input.rhs)

    app.add_middleware(
        CORSMiddleware,
        allow_credentials=True,
        allow_headers=("*"),
        allow_methods=("*"),
        allow_origins=("*"),
    )

    run(app, host="0.0.0.0", port=8080)


class StatefulAdditionApp(fal.App, keep_alive=300, max_concurrency=1):
    machine_type = "S"

    async def setup(self):
        self.counter = 0

    @fal.endpoint("/reset")
    def reset(self) -> Output:
        self.counter = 0
        return Output(result=self.counter)

    @fal.endpoint("/increment")
    def increment(self, input: StatefulInput) -> Output:
        self.counter += input.value
        return Output(result=self.counter)

    @fal.endpoint("/decrement")
    def decrement(self, input: StatefulInput) -> Output:
        self.counter -= input.value
        return Output(result=self.counter)


class ExceptionApp(fal.App, keep_alive=300, max_concurrency=1):
    machine_type = "XS"

    @fal.endpoint("/fail")
    def fail(self) -> Output:
        raise Exception("this app is designed to fail!")

    @fal.endpoint("/app-exception")
    def app_exception(self) -> Output:
        raise AppException(message="this app is designed to fail", status_code=401)

    @fal.endpoint("/field-exception")
    def field_exception(self, input: Input) -> Output:
        raise FieldException(
            field="rhs",
            message="rhs must be an integer",
        )

    @fal.endpoint("/cuda-exception")
    def cuda_exception(self) -> Output:
        # mimicking error message from PyTorch (https://github.com/pytorch/pytorch/blob/6c65fd03942415b68040e102c44cf5109d2d851e/c10/cuda/CUDACachingAllocator.cpp#L1234C12-L1234C30)
        raise RuntimeError("CUDA out of memory")


class CancellableApp(fal.App, keep_alive=300, max_concurrency=1):
    task = None

    @fal.endpoint("/")
    async def sleep(self, input: Input) -> Output:
        self.task = asyncio.create_task(asyncio.sleep(input.wait_time))
        try:
            await self.task
        except asyncio.CancelledError:
            print("Task was cancelled")
            if not self.task.done():
                self.task.cancel()
                with suppress(Exception):
                    await self.task

            raise RequestCancelledException("Request cancelled by the client.")

        return Output(result=input.lhs + input.rhs)

    @fal.endpoint("/cancel")
    async def cancel_handler(self) -> Output:
        if self.task:
            self.task.cancel()
            with suppress(BaseException):
                await self.task
            self.task = None

        return Output(result=0)


class RTInput(BaseModel):
    prompt: str

    def can_batch(
        self,
        other: "RTInput",
        current_batch_size: int = 1,
    ) -> bool:
        return "don't batch" not in other.prompt


class RTOutput(BaseModel):
    text: str


class RTOutputs(BaseModel):
    texts: list[str]


class RealtimeApp(fal.App, keep_alive=300, max_concurrency=1):
    machine_type = "S"

    @fal.endpoint("/")
    def generate(self, input: RTInput) -> RTOutput:
        return RTOutput(text=input.prompt)

    @fal.endpoint("/ws", is_websocket=True)
    async def generate_ws(self, websocket: WebSocket) -> None:
        await websocket.accept()
        for _ in range(3):
            await websocket.send_json({"message": "Hello world!"})
        await websocket.close()

    @fal.realtime("/realtime")
    def generate_rt(self, input: RTInput) -> RTOutput:
        return RTOutput(text=input.prompt)

    @fal.realtime("/realtime/batched", buffering=10, max_batch_size=4)
    def generate_rt_batched(self, input: RTInput, *inputs: RTInput) -> RTOutputs:
        time.sleep(2)  # fixed cost
        return RTOutputs(texts=[input.prompt] + [i.prompt for i in inputs])


@pytest.fixture(scope="module")
def aliased_app() -> Generator[tuple[str, str], None, None]:
    # Create a temporary app, register it, and return the ID of it.

    import uuid

    app_alias = str(uuid.uuid4()) + "-alias"
    app_revision = addition_app.host.register(
        func=addition_app.func,
        options=addition_app.options,
        # random enough
        application_name=app_alias,
        application_auth_mode="private",
    )
    yield app_revision, app_alias  # type: ignore


@pytest.fixture(scope="module")
def test_app():
    # Create a temporary app, register it, and return the ID of it.

    app_revision = addition_app.host.register(
        func=addition_app.func,
        options=addition_app.options,
    )
    user = _get_user()
    yield f"{user.user_id}/{app_revision}"


@pytest.fixture(scope="module")
def test_nomad_app():
    # Create a temporary app, register it, and return the ID of it.

    app_revision = nomad_addition_app.host.register(
        func=nomad_addition_app.func,
        options=nomad_addition_app.options,
    )
    user = _get_user()
    yield f"{user.user_id}/{app_revision}"


@pytest.fixture(scope="module")
def test_container_app():
    # Create a temporary app, register it, and return the ID of it.

    app_revision = container_addition_app.host.register(
        func=container_addition_app.func,
        options=container_addition_app.options,
    )
    user = _get_user()
    yield f"{user.user_id}/{app_revision}"


@pytest.fixture(scope="module")
def test_fastapi_app():
    # Create a temporary app, register it, and return the ID of it.

    app_revision = calculator_app.host.register(
        func=calculator_app.func,
        options=calculator_app.options,
    )
    user = _get_user()
    yield f"{user.user_id}/{app_revision}"


@pytest.fixture(scope="module")
def test_stateful_app():
    # Create a temporary app, register it, and return the ID of it.

    app = fal.wrap_app(StatefulAdditionApp)
    app_revision = app.host.register(
        func=app.func,
        options=app.options,
    )
    user = _get_user()
    yield f"{user.user_id}/{app_revision}"


@pytest.fixture(scope="module")
def test_cancellable_app():
    # Create a temporary app, register it, and return the ID of it.

    app = fal.wrap_app(CancellableApp)
    app_revision = app.host.register(
        func=app.func,
        options=app.options,
        application_auth_mode="public",
    )
    user = _get_user()
    yield f"{user.user_id}/{app_revision}"


@pytest.fixture(scope="module")
def test_exception_app():
    # Create a temporary app, register it, and return the ID of it.
    with AppClient.connect(ExceptionApp) as client:
        yield client


@pytest.fixture(scope="module")
def test_realtime_app():
    # Create a temporary app, register it, and return the ID of it.

    app = fal.wrap_app(RealtimeApp)
    app_revision = app.host.register(
        func=app.func,
        options=app.options,
        application_auth_mode="public",
    )
    user = _get_user()
    yield f"{user.user_id}/{app_revision}"


def test_app_client(test_app: str, test_nomad_app: str):
    response = apps.run(test_app, arguments={"lhs": 1, "rhs": 2})
    assert response["result"] == 3

    response = apps.run(test_app, arguments={"lhs": 2, "rhs": 3, "wait_time": 1})
    assert response["result"] == 5

    response = apps.run(test_nomad_app, arguments={"lhs": 1, "rhs": 2})
    assert response["result"] == 3

    response = apps.run(test_nomad_app, arguments={"lhs": 2, "rhs": 3, "wait_time": 1})
    assert response["result"] == 5


def test_app_client_old_format(test_app: str):
    assert test_app.count("/") == 1, "Test app should be in new format"
    old_format = test_app.replace("/", "-")
    assert test_app.count("-") + 1 == old_format.count(
        "-"
    ), "Old format should have one more hyphen"

    response = apps.run(old_format, arguments={"lhs": 1, "rhs": 2})
    assert response["result"] == 3


def test_app_client_path_included_in_app_id(test_stateful_app: str):
    response = apps.run(test_stateful_app + "/reset", arguments={})
    assert response["result"] == 0

    response = apps.run(test_stateful_app + "/increment", arguments={"value": 3})
    assert response["result"] == 3

    # if put in path we do not need to prefix with /
    response = apps.run(test_stateful_app, arguments={"value": 3}, path="increment")
    assert response["result"] == 6


def test_stateful_app_client(test_stateful_app: str):
    response = apps.run(test_stateful_app, arguments={}, path="/reset")
    assert response["result"] == 0

    response = apps.run(test_stateful_app, arguments={"value": 1}, path="/increment")
    assert response["result"] == 1

    response = apps.run(test_stateful_app, arguments={"value": 2}, path="/increment")
    assert response["result"] == 3

    response = apps.run(test_stateful_app, arguments={"value": 1}, path="/decrement")
    assert response["result"] == 2

    response = apps.run(test_stateful_app, arguments={"value": 2}, path="/decrement")
    assert response["result"] == 0


def test_app_cancellation(test_app: str, test_cancellable_app: str):
    request_handle = apps.submit(
        test_cancellable_app, arguments={"lhs": 1, "rhs": 2, "wait_time": 10}
    )
    # enough time for it to start
    time.sleep(8)
    request_handle.cancel()

    # should still finish successfully and return 499
    with pytest.raises(HTTPStatusError) as e:
        request_handle.get()
    assert e.value.response.status_code == 499

    # normal app should just ignore the cancellation
    request_handle = apps.submit(
        test_app, arguments={"lhs": 1, "rhs": 2, "wait_time": 10}
    )

    # enough time for it to start
    time.sleep(8)
    request_handle.cancel()

    response = request_handle.get()
    assert response == {"result": 3}


@pytest.mark.flaky(max_runs=3)
def test_app_client_async(test_app: str):
    request_handle = apps.submit(test_app, arguments={"lhs": 1, "rhs": 2})
    assert request_handle.get() == {"result": 3}

    request_handle = apps.submit(
        test_app, arguments={"lhs": 2, "rhs": 3, "wait_time": 5}
    )

    for event in request_handle.iter_events(logs=True):
        assert isinstance(event, (apps.Queued, apps.InProgress))
        if isinstance(event, apps.InProgress) and event.logs:
            logs = [log["message"] for log in event.logs]
            assert "sleeping..." in logs
        elif isinstance(event, apps.Queued):
            assert event.position == 0

    status = request_handle.status(logs=True)
    assert isinstance(status, apps.Completed)
    assert status.logs, "Logs missing from Completed status"

    # It is safe to use fetch_result when we know for a fact the request itself
    # is completed.
    result = request_handle.fetch_result()

    # .get() can still be used and will return the same value
    result_alternative = request_handle.get()
    assert result == result_alternative
    assert result == {"result": 5}


def test_app_openapi_spec_metadata(test_app: str, request: pytest.FixtureRequest):
    user_id, _, app_id = test_app.partition("/")
    res = app_metadata.sync_detailed(
        app_alias_or_id=app_id, app_user_id=user_id, client=REST_CLIENT
    )

    assert res.status_code == 200, f"Failed to fetch metadata for app {test_app}"
    assert res.parsed, f"Failed to parse metadata for app {test_app}"

    metadata = res.parsed.to_dict()
    assert "openapi" in metadata, f"openapi key missing from metadata {metadata}"
    openapi_spec: dict = metadata["openapi"]
    for key in ["openapi", "info", "paths", "components"]:
        assert key in openapi_spec, f"{key} key missing from openapi {openapi_spec}"


def test_app_no_serve_spec_metadata(
    test_fastapi_app: str, request: pytest.FixtureRequest
):
    # We do not store the openapi spec for apps that do not use serve=True
    user_id, _, app_id = test_fastapi_app.partition("/")
    res = app_metadata.sync_detailed(
        app_alias_or_id=app_id, app_user_id=user_id, client=REST_CLIENT
    )

    assert (
        res.status_code == 200
    ), f"Failed to fetch metadata for app {test_fastapi_app}"
    assert res.parsed, f"Failed to parse metadata for app {test_fastapi_app}"

    metadata = res.parsed.to_dict()
    assert (
        "openapi" not in metadata
    ), f"openapi should not be present in metadata {metadata}"


def test_404_response(test_app: str, request: pytest.FixtureRequest):
    with pytest.raises(HTTPStatusError, match="Path /other not found"):
        apps.run(test_app, path="/other", arguments={"lhs": 1, "rhs": 2})


def test_app_update_app(aliased_app: tuple[str, str]):
    app_revision, app_alias = aliased_app

    host: api.FalServerlessHost = addition_app.host  # type: ignore
    with host._connection as client:
        # Get the registered values
        res = client.list_aliases()
        found = next(filter(lambda alias: alias.alias == app_alias, res), None)
        assert found, f"Could not find app {app_alias} in {res}"
        assert found.revision == app_revision

    with host._connection as client:
        new_keep_alive = found.keep_alive + 1
        new_max_concurrency = found.max_concurrency + 1
        new_max_multiplexing = found.max_multiplexing + 1

        res = client.update_application(
            application_name=app_alias,
            keep_alive=new_keep_alive,
            max_concurrency=new_max_concurrency,
            max_multiplexing=new_max_multiplexing,
        )
        assert res.alias == app_alias
        assert res.keep_alive == new_keep_alive
        assert res.max_concurrency == new_max_concurrency
        assert res.max_multiplexing == new_max_multiplexing

    with host._connection as client:
        new_keep_alive = new_keep_alive + 1
        res = client.update_application(
            application_name=app_alias,
            keep_alive=new_keep_alive,
        )
        assert res.alias == app_alias
        assert res.keep_alive == new_keep_alive
        assert res.max_concurrency == new_max_concurrency
        assert res.max_multiplexing == new_max_multiplexing

    with host._connection as client:
        new_max_concurrency = new_max_concurrency - 1
        res = client.update_application(
            application_name=app_alias,
            max_concurrency=new_max_concurrency,
        )
        assert res.alias == app_alias
        assert res.keep_alive == new_keep_alive
        assert res.max_concurrency == new_max_concurrency
        assert res.max_multiplexing == new_max_multiplexing


def test_app_set_delete_alias(aliased_app: tuple[str, str]):
    app_revision, app_alias = aliased_app

    host: api.FalServerlessHost = addition_app.host  # type: ignore

    with host._connection as client:
        # Get the registered values
        res = client.list_aliases()
        found = next(filter(lambda alias: alias.alias == app_alias, res), None)
        assert found, f"Could not find app {app_alias} in {res}"
        assert found.revision == app_revision
        assert found.auth_mode == "private"

    with host._connection as client:
        # Get the registered values
        res = client.create_alias(app_alias, app_revision, "public")

    with host._connection as client:
        # Get the registered values
        res = client.list_aliases()
        found = next(filter(lambda alias: alias.alias == app_alias, res), None)
        assert found, f"Could not find app {app_alias} in {res}"
        assert found.revision == app_revision
        assert found.auth_mode == "public"

    with host._connection as client:
        res = client.delete_alias(alias=app_alias)
        assert res == app_revision

    with host._connection as client:
        # Get the registered values
        res = client.list_aliases()
        found = next(filter(lambda alias: alias.alias == app_alias, res), None)
        assert not found, f"Found app {app_alias} in {res} after deletion"


@pytest.mark.flaky(max_runs=3)
def test_realtime_connection(test_realtime_app):
    response = apps.run(test_realtime_app, arguments={"prompt": "a cat"})
    assert response["text"] == "a cat"

    with apps._connect(test_realtime_app) as connection:
        for _ in range(3):
            response = connection.run({"prompt": "a cat"})
            assert response["text"] == "a cat"

    with apps._connect(test_realtime_app, path="/realtime/batched") as connection:
        connection.send({"prompt": "keep busy"})
        time.sleep(0.1)

        for prompt in range(10):
            connection.send({"prompt": str(prompt)})

        assert connection.recv()["texts"] == ["keep busy"]

        received_prompts = set()
        batch_sizes = []
        while len(received_prompts) < 10:
            response = connection.recv()
            received_prompts.update(response["texts"])
            batch_sizes.append(len(response["texts"]))

        assert len(received_prompts) == 10
        assert batch_sizes == [4, 4, 2]


@contextmanager
def delete_workflow_on_exit(client: httpx.Client, workflow_url: str):
    try:
        yield
    finally:
        client.delete(workflow_url)


def test_workflows(test_app: str):
    workflow = Workflow(
        name="test_workflow_" + secrets.token_hex(),
        input_schema={},
        output_schema={},
    )
    # (lhs + rhs) + (lhs + rhs)
    lhs = workflow.run(
        test_app,
        input={
            "lhs": workflow.input.lhs,
            "rhs": workflow.input.rhs,
        },
    )
    rhs = workflow.run(
        test_app,
        input={
            "lhs": workflow.input.lhs,
            "rhs": workflow.input.rhs,
        },
    )
    out = workflow.run(
        test_app,
        input={
            "lhs": lhs.result,
            "rhs": rhs.result,
        },
    )
    workflow.set_output({"result": out.result})
    workflow_id = workflow.publish(title="Test Workflow", is_public=False)

    # Test the underlying app
    data = fal.apps.run(test_app, arguments={"lhs": 2, "rhs": 3})
    assert data["result"] == 5

    with httpx.Client(
        base_url=REST_CLIENT.base_url,
        headers=REST_CLIENT.get_headers(),
        timeout=300,
    ) as client:
        with delete_workflow_on_exit(
            client, REST_CLIENT.base_url + "/workflows/" + workflow_id
        ):
            data = fal.apps.run(
                "workflows/" + workflow_id, arguments={"lhs": 2, "rhs": 3}
            )
            assert data["result"] == 10


def test_traceback_logs(test_exception_app: AppClient):
    date = datetime.utcnow().isoformat()

    with pytest.raises(HTTPStatusError):
        test_exception_app.fail({})

    with httpx.Client(
        base_url=REST_CLIENT.base_url,
        headers=REST_CLIENT.get_headers(),
        timeout=300,
    ) as client:
        # Give some time for logs to propagate through the logging subsystem.
        time.sleep(5)
        response = client.get(
            REST_CLIENT.base_url + f"/logs/?traceback=true&since={date}"
        )
        for log in json.loads(response.text):
            assert log["message"].count("\n") > 1, "Logs are multi-line"
            assert '{"traceback":' not in log["message"], "Logs are not JSON-wrapped"
            assert (
                "this app is designed to fail" in log["message"]
            ), "Logs contain the traceback message"


def test_app_exceptions(test_exception_app: AppClient):
    with pytest.raises(HTTPStatusError) as app_exc:
        test_exception_app.app_exception({})

    assert app_exc.value.response.status_code == 401

    with pytest.raises(HTTPStatusError) as field_exc:
        test_exception_app.field_exception({"lhs": 1, "rhs": "2"})

    assert field_exc.value.response.status_code == 422

    with pytest.raises(HTTPStatusError) as cuda_exc:
        test_exception_app.cuda_exception({})

    assert cuda_exc.value.response.status_code == _CUDA_OOM_STATUS_CODE
    assert cuda_exc.value.response.json()["detail"] == _CUDA_OOM_MESSAGE



from typing import TYPE_CHECKING, Any, Dict, List, Type, TypeVar, Union

import attr

from ..types import UNSET, Unset

if TYPE_CHECKING:
    from ..models.comfy_workflow_schema_extra_data import ComfyWorkflowSchemaExtraData
    from ..models.comfy_workflow_schema_fal_inputs import ComfyWorkflowSchemaFalInputs
    from ..models.comfy_workflow_schema_fal_inputs_dev_info import ComfyWorkflowSchemaFalInputsDevInfo
    from ..models.comfy_workflow_schema_prompt import ComfyWorkflowSchemaPrompt


T = TypeVar("T", bound="ComfyWorkflowSchema")


@attr.s(auto_attribs=True)
class ComfyWorkflowSchema:
    """
    Attributes:
        prompt (ComfyWorkflowSchemaPrompt):
        extra_data (Union[Unset, ComfyWorkflowSchemaExtraData]):
        fal_inputs_dev_info (Union[Unset, ComfyWorkflowSchemaFalInputsDevInfo]):
        fal_inputs (Union[Unset, ComfyWorkflowSchemaFalInputs]):
    """

    prompt: "ComfyWorkflowSchemaPrompt"
    extra_data: Union[Unset, "ComfyWorkflowSchemaExtraData"] = UNSET
    fal_inputs_dev_info: Union[Unset, "ComfyWorkflowSchemaFalInputsDevInfo"] = UNSET
    fal_inputs: Union[Unset, "ComfyWorkflowSchemaFalInputs"] = UNSET
    additional_properties: Dict[str, Any] = attr.ib(init=False, factory=dict)

    def to_dict(self) -> Dict[str, Any]:
        prompt = self.prompt.to_dict()

        extra_data: Union[Unset, Dict[str, Any]] = UNSET
        if not isinstance(self.extra_data, Unset):
            extra_data = self.extra_data.to_dict()

        fal_inputs_dev_info: Union[Unset, Dict[str, Any]] = UNSET
        if not isinstance(self.fal_inputs_dev_info, Unset):
            fal_inputs_dev_info = self.fal_inputs_dev_info.to_dict()

        fal_inputs: Union[Unset, Dict[str, Any]] = UNSET
        if not isinstance(self.fal_inputs, Unset):
            fal_inputs = self.fal_inputs.to_dict()

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "prompt": prompt,
            }
        )
        if extra_data is not UNSET:
            field_dict["extra_data"] = extra_data
        if fal_inputs_dev_info is not UNSET:
            field_dict["fal_inputs_dev_info"] = fal_inputs_dev_info
        if fal_inputs is not UNSET:
            field_dict["fal_inputs"] = fal_inputs

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        from ..models.comfy_workflow_schema_extra_data import ComfyWorkflowSchemaExtraData
        from ..models.comfy_workflow_schema_fal_inputs import ComfyWorkflowSchemaFalInputs
        from ..models.comfy_workflow_schema_fal_inputs_dev_info import ComfyWorkflowSchemaFalInputsDevInfo
        from ..models.comfy_workflow_schema_prompt import ComfyWorkflowSchemaPrompt

        d = src_dict.copy()
        prompt = ComfyWorkflowSchemaPrompt.from_dict(d.pop("prompt"))

        _extra_data = d.pop("extra_data", UNSET)
        extra_data: Union[Unset, ComfyWorkflowSchemaExtraData]
        if isinstance(_extra_data, Unset):
            extra_data = UNSET
        else:
            extra_data = ComfyWorkflowSchemaExtraData.from_dict(_extra_data)

        _fal_inputs_dev_info = d.pop("fal_inputs_dev_info", UNSET)
        fal_inputs_dev_info: Union[Unset, ComfyWorkflowSchemaFalInputsDevInfo]
        if isinstance(_fal_inputs_dev_info, Unset):
            fal_inputs_dev_info = UNSET
        else:
            fal_inputs_dev_info = ComfyWorkflowSchemaFalInputsDevInfo.from_dict(_fal_inputs_dev_info)

        _fal_inputs = d.pop("fal_inputs", UNSET)
        fal_inputs: Union[Unset, ComfyWorkflowSchemaFalInputs]
        if isinstance(_fal_inputs, Unset):
            fal_inputs = UNSET
        else:
            fal_inputs = ComfyWorkflowSchemaFalInputs.from_dict(_fal_inputs)

        comfy_workflow_schema = cls(
            prompt=prompt,
            extra_data=extra_data,
            fal_inputs_dev_info=fal_inputs_dev_info,
            fal_inputs=fal_inputs,
        )

        comfy_workflow_schema.additional_properties = d
        return comfy_workflow_schema

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from __future__ import annotations

import json
import os
import posixpath
import uuid
from dataclasses import dataclass
from io import BytesIO

from fal.toolkit.file.types import FileData, FileRepository

DEFAULT_URL_TIMEOUT = 60 * 15  # 15 minutes


@dataclass
class R2Repository(FileRepository):
    bucket_name: str = "fal_file_storage"
    url_expiration: int = DEFAULT_URL_TIMEOUT
    r2_account_json: str | None = None
    key: str = ""

    _storage_client = None
    _bucket = None

    def __post_init__(self):
        import boto3
        from botocore.client import Config

        r2_account_json = self.r2_account_json
        if r2_account_json is None:
            r2_account_json = os.environ.get("R2_CREDS_JSON")
            if r2_account_json is None:
                raise Exception("R2_CREDS_JSON environment secret is not set")

        r2_account_info = json.loads(r2_account_json)
        account_id = r2_account_info["ACCOUNT_ID"]
        access_key_id = r2_account_info["ACCESS_KEY_ID"]
        secret_access_key = r2_account_info["SECRET_ACCESS_KEY"]

        self._s3_client = boto3.client(
            "s3",
            endpoint_url=f"https://{account_id}.r2.cloudflarestorage.com",
            aws_access_key_id=access_key_id,
            aws_secret_access_key=secret_access_key,
            config=Config(signature_version="s3v4"),
        )
        self._s3_resource = boto3.resource(
            "s3",
            endpoint_url=f"https://{account_id}.r2.cloudflarestorage.com",
            aws_access_key_id=access_key_id,
            aws_secret_access_key=secret_access_key,
        )

        self._bucket = self._s3_resource.Bucket(self.bucket_name)

    @property
    def storage_client(self):
        if self._s3_resource is None:
            raise Exception("S3 Resource is not initialized")

        return self._s3_resource

    @property
    def bucket(self):
        if self._bucket is None:
            raise Exception("S3 bucket is not initialized")

        return self._bucket

    def save(self, data: FileData) -> str:
        destination_path = posixpath.join(
            self.key,
            f"{uuid.uuid4().hex}_{data.file_name}",
        )

        s3_object = self.bucket.Object(destination_path)
        s3_object.upload_fileobj(
            BytesIO(data.data), ExtraArgs={"ContentType": data.content_type}
        )

        public_url = self._s3_client.generate_presigned_url(
            ClientMethod="get_object",
            Params={"Bucket": self.bucket_name, "Key": destination_path},
            ExpiresIn=self.url_expiration,
        )
        return public_url



from typing import TYPE_CHECKING, Any, Dict, List, Type, TypeVar, Union

import attr

from ..types import UNSET, Unset

if TYPE_CHECKING:
    from ..models.comfy_workflow_schema import ComfyWorkflowSchema


T = TypeVar("T", bound="TypedComfyWorkflowUpdate")


@attr.s(auto_attribs=True)
class TypedComfyWorkflowUpdate:
    """
    Attributes:
        title (Union[Unset, str]):
        workflow (Union[Unset, ComfyWorkflowSchema]):
        is_public (Union[Unset, bool]):
        name (Union[Unset, str]):
    """

    title: Union[Unset, str] = UNSET
    workflow: Union[Unset, "ComfyWorkflowSchema"] = UNSET
    is_public: Union[Unset, bool] = UNSET
    name: Union[Unset, str] = UNSET
    additional_properties: Dict[str, Any] = attr.ib(init=False, factory=dict)

    def to_dict(self) -> Dict[str, Any]:
        title = self.title
        workflow: Union[Unset, Dict[str, Any]] = UNSET
        if not isinstance(self.workflow, Unset):
            workflow = self.workflow.to_dict()

        is_public = self.is_public
        name = self.name

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update({})
        if title is not UNSET:
            field_dict["title"] = title
        if workflow is not UNSET:
            field_dict["workflow"] = workflow
        if is_public is not UNSET:
            field_dict["is_public"] = is_public
        if name is not UNSET:
            field_dict["name"] = name

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        from ..models.comfy_workflow_schema import ComfyWorkflowSchema

        d = src_dict.copy()
        title = d.pop("title", UNSET)

        _workflow = d.pop("workflow", UNSET)
        workflow: Union[Unset, ComfyWorkflowSchema]
        if isinstance(_workflow, Unset):
            workflow = UNSET
        else:
            workflow = ComfyWorkflowSchema.from_dict(_workflow)

        is_public = d.pop("is_public", UNSET)

        name = d.pop("name", UNSET)

        typed_comfy_workflow_update = cls(
            title=title,
            workflow=workflow,
            is_public=is_public,
            name=name,
        )

        typed_comfy_workflow_update.additional_properties = d
        return typed_comfy_workflow_update

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from __future__ import annotations

import dataclasses
import json
import math
import os
from base64 import b64encode
from dataclasses import dataclass
from pathlib import Path
from urllib.error import HTTPError
from urllib.request import Request, urlopen

from fal.auth import key_credentials
from fal.toolkit.exceptions import FileUploadException
from fal.toolkit.file.types import FileData, FileRepository

_FAL_CDN = "https://fal.media"


@dataclass
class ObjectLifecyclePreference:
    expriation_duration_seconds: int


GLOBAL_LIFECYCLE_PREFERENCE = ObjectLifecyclePreference(
    expriation_duration_seconds=86400
)


@dataclass
class FalFileRepositoryBase(FileRepository):
    def _save(self, file: FileData, storage_type: str) -> str:
        key_creds = key_credentials()
        if not key_creds:
            raise FileUploadException("FAL_KEY must be set")

        key_id, key_secret = key_creds
        headers = {
            "Authorization": f"Key {key_id}:{key_secret}",
            "Accept": "application/json",
            "Content-Type": "application/json",
        }

        grpc_host = os.environ.get("FAL_HOST", "api.alpha.fal.ai")
        rest_host = grpc_host.replace("api", "rest", 1)
        storage_url = (
            f"https://{rest_host}/storage/upload/initiate?storage_type={storage_type}"
        )

        try:
            req = Request(
                storage_url,
                data=json.dumps(
                    {
                        "file_name": file.file_name,
                        "content_type": file.content_type,
                    }
                ).encode(),
                headers=headers,
                method="POST",
            )
            with urlopen(req) as response:
                result = json.load(response)

            upload_url = result["upload_url"]
            self._upload_file(upload_url, file)

            return result["file_url"]
        except HTTPError as e:
            raise FileUploadException(
                f"Error initiating upload. Status {e.status}: {e.reason}"
            )

    def _upload_file(self, upload_url: str, file: FileData):
        req = Request(
            upload_url,
            method="PUT",
            data=file.data,
            headers={"Content-Type": file.content_type},
        )

        with urlopen(req):
            return


@dataclass
class FalFileRepository(FalFileRepositoryBase):
    def save(self, file: FileData) -> str:
        return self._save(file, "gcs")


class MultipartUpload:
    MULTIPART_THRESHOLD = 100 * 1024 * 1024
    MULTIPART_CHUNK_SIZE = 100 * 1024 * 1024
    MULTIPART_MAX_CONCURRENCY = 10

    def __init__(
        self,
        file_path: str | Path,
        chunk_size: int | None = None,
        content_type: str | None = None,
        max_concurrency: int | None = None,
    ) -> None:
        self.file_path = file_path
        self.chunk_size = chunk_size or self.MULTIPART_CHUNK_SIZE
        self.content_type = content_type or "application/octet-stream"
        self.max_concurrency = max_concurrency or self.MULTIPART_MAX_CONCURRENCY

        self._parts: list[dict] = []

        key_creds = key_credentials()
        if not key_creds:
            raise FileUploadException("FAL_KEY must be set")

        key_id, key_secret = key_creds

        self._auth_headers = {
            "Authorization": f"Key {key_id}:{key_secret}",
        }
        grpc_host = os.environ.get("FAL_HOST", "api.alpha.fal.ai")
        rest_host = grpc_host.replace("api", "rest", 1)
        self._storage_upload_url = f"https://{rest_host}/storage/upload"

    def create(self):
        try:
            req = Request(
                f"{self._storage_upload_url}/initiate-multipart",
                method="POST",
                headers={
                    **self._auth_headers,
                    "Accept": "application/json",
                    "Content-Type": "application/json",
                },
                data=json.dumps(
                    {
                        "file_name": os.path.basename(self.file_path),
                        "content_type": self.content_type,
                    }
                ).encode(),
            )
            with urlopen(req) as response:
                result = json.load(response)
                self._upload_id = result["upload_id"]
                self._file_url = result["file_url"]
        except HTTPError as exc:
            raise FileUploadException(
                f"Error initiating upload. Status {exc.status}: {exc.reason}"
            )

    def _upload_part(self, url: str, part_number: int) -> dict:
        with open(self.file_path, "rb") as f:
            start = (part_number - 1) * self.chunk_size
            f.seek(start)
            data = f.read(self.chunk_size)
            req = Request(
                url,
                method="PUT",
                headers={"Content-Type": self.content_type},
                data=data,
            )

            try:
                with urlopen(req) as resp:
                    return {
                        "part_number": part_number,
                        "etag": resp.headers["ETag"],
                    }
            except HTTPError as exc:
                raise FileUploadException(
                    f"Error uploading part {part_number} to {url}. "
                    f"Status {exc.status}: {exc.reason}"
                )

    def upload(self) -> None:
        import concurrent.futures

        parts = math.ceil(os.path.getsize(self.file_path) / self.chunk_size)
        with concurrent.futures.ThreadPoolExecutor(
            max_workers=self.max_concurrency
        ) as executor:
            futures = []
            for part_number in range(1, parts + 1):
                upload_url = (
                    f"{self._file_url}?upload_id={self._upload_id}"
                    f"&part_number={part_number}"
                )
                futures.append(
                    executor.submit(self._upload_part, upload_url, part_number)
                )

            for future in concurrent.futures.as_completed(futures):
                entry = future.result()
                self._parts.append(entry)

    def complete(self):
        url = f"{self._file_url}?upload_id={self._upload_id}"
        try:
            req = Request(
                url,
                method="POST",
                headers={
                    "Accept": "application/json",
                    "Content-Type": "application/json",
                },
                data=json.dumps({"parts": self._parts}).encode(),
            )
            with urlopen(req):
                pass
        except HTTPError as e:
            raise FileUploadException(
                f"Error completing upload {url}. Status {e.status}: {e.reason}"
            )

        return self._file_url


@dataclass
class FalFileRepositoryV2(FalFileRepositoryBase):
    def save(self, file: FileData) -> str:
        return self._save(file, "fal-cdn")

    def _save_multipart(
        self,
        file_path: str | Path,
        chunk_size: int | None = None,
        content_type: str | None = None,
        max_concurrency: int | None = None,
    ) -> str:
        multipart = MultipartUpload(
            file_path,
            chunk_size=chunk_size,
            content_type=content_type,
            max_concurrency=max_concurrency,
        )
        multipart.create()
        multipart.upload()
        return multipart.complete()

    def save_file(
        self,
        file_path: str | Path,
        content_type: str,
        multipart: bool | None = None,
        multipart_threshold: int | None = None,
        multipart_chunk_size: int | None = None,
        multipart_max_concurrency: int | None = None,
    ) -> tuple[str, FileData | None]:
        if multipart is None:
            threshold = multipart_threshold or MultipartUpload.MULTIPART_THRESHOLD
            multipart = os.path.getsize(file_path) > threshold

        if multipart:
            url = self._save_multipart(
                file_path,
                chunk_size=multipart_chunk_size,
                content_type=content_type,
                max_concurrency=multipart_max_concurrency,
            )
            data = None
        else:
            with open(file_path, "rb") as f:
                data = FileData(
                    f.read(),
                    content_type=content_type,
                    file_name=os.path.basename(file_path),
                )
            url = self.save(data)

        return url, data


@dataclass
class InMemoryRepository(FileRepository):
    def save(
        self,
        file: FileData,
    ) -> str:
        return f'data:{file.content_type};base64,{b64encode(file.data).decode("utf-8")}'


@dataclass
class FalCDNFileRepository(FileRepository):
    def save(
        self,
        file: FileData,
    ) -> str:
        headers = {
            **self.auth_headers,
            "Accept": "application/json",
            "Content-Type": file.content_type,
            "X-Fal-File-Name": file.file_name,
            "X-Fal-Object-Lifecycle-Preference": json.dumps(
                dataclasses.asdict(GLOBAL_LIFECYCLE_PREFERENCE)
            ),
        }
        url = os.getenv("FAL_CDN_HOST", _FAL_CDN) + "/files/upload"
        request = Request(url, headers=headers, method="POST", data=file.data)
        try:
            with urlopen(request) as response:
                result = json.load(response)
        except HTTPError as e:
            raise FileUploadException(
                f"Error initiating upload. Status {e.status}: {e.reason}"
            )

        access_url = result["access_url"]
        return access_url

    @property
    def auth_headers(self) -> dict[str, str]:
        key_creds = key_credentials()
        if not key_creds:
            raise FileUploadException("FAL_KEY must be set")

        key_id, key_secret = key_creds
        return {
            "Authorization": f"Bearer {key_id}:{key_secret}",
            "User-Agent": "fal/0.1.0",
        }



from __future__ import annotations

import shutil
from pathlib import Path
from tempfile import NamedTemporaryFile, mkdtemp
from typing import Any, Callable, Optional
from urllib.parse import urlparse
from zipfile import ZipFile

import pydantic

# https://github.com/pydantic/pydantic/pull/2573
if not hasattr(pydantic, "__version__") or pydantic.__version__.startswith("1."):
    IS_PYDANTIC_V2 = False
else:
    from pydantic import GetCoreSchemaHandler
    from pydantic_core import CoreSchema, core_schema

    IS_PYDANTIC_V2 = True

from pydantic import BaseModel, Field

from fal.toolkit.file.providers.fal import (
    FalCDNFileRepository,
    FalFileRepository,
    FalFileRepositoryV2,
    InMemoryRepository,
)
from fal.toolkit.file.providers.gcp import GoogleStorageRepository
from fal.toolkit.file.providers.r2 import R2Repository
from fal.toolkit.file.types import FileData, FileRepository, RepositoryId
from fal.toolkit.utils.download_utils import download_file

FileRepositoryFactory = Callable[[], FileRepository]

BUILT_IN_REPOSITORIES: dict[RepositoryId, FileRepositoryFactory] = {
    "fal": lambda: FalFileRepository(),
    "fal_v2": lambda: FalFileRepositoryV2(),
    "in_memory": lambda: InMemoryRepository(),
    "gcp_storage": lambda: GoogleStorageRepository(),
    "r2": lambda: R2Repository(),
    "cdn": lambda: FalCDNFileRepository(),
}


def get_builtin_repository(id: RepositoryId) -> FileRepository:
    if id not in BUILT_IN_REPOSITORIES.keys():
        raise ValueError(f'"{id}" is not a valid built-in file repository')
    return BUILT_IN_REPOSITORIES[id]()


get_builtin_repository.__module__ = "__main__"

DEFAULT_REPOSITORY: FileRepository | RepositoryId = "fal"


class File(BaseModel):
    # public properties
    url: str = Field(
        description="The URL where the file can be downloaded from.",
    )
    content_type: Optional[str] = Field(
        None,
        description="The mime type of the file.",
        examples=["image/png"],
    )
    file_name: Optional[str] = Field(
        None,
        description="The name of the file. It will be auto-generated if not provided.",
        examples=["z9RV14K95DvU.png"],
    )
    file_size: Optional[int] = Field(
        None, description="The size of the file in bytes.", examples=[4404019]
    )
    file_data: Optional[bytes] = Field(
        None,
        description="File data",
        exclude=True,
        repr=False,
    )

    # Pydantic custom validator for input type conversion
    if IS_PYDANTIC_V2:

        @classmethod
        def __get_pydantic_core_schema__(
            cls, source_type: Any, handler: GetCoreSchemaHandler
        ) -> CoreSchema:
            return core_schema.no_info_before_validator_function(
                cls.__convert_from_str,
                handler(source_type),
            )

    else:

        @classmethod
        def __get_validators__(cls):
            yield cls.__convert_from_str

    @classmethod
    def __convert_from_str(cls, value: Any):
        if isinstance(value, str):
            parsed_url = urlparse(value)
            if parsed_url.scheme not in ["http", "https", "data"]:
                raise ValueError("value must be a valid URL")
            return cls._from_url(parsed_url.geturl())

        return value

    @classmethod
    def _from_url(
        cls,
        url: str,
    ) -> File:
        return cls(
            url=url,
            content_type=None,
            file_name=None,
            repository=DEFAULT_REPOSITORY,
        )

    @classmethod
    def from_bytes(
        cls,
        data: bytes,
        content_type: Optional[str] = None,
        file_name: Optional[str] = None,
        repository: FileRepository | RepositoryId = DEFAULT_REPOSITORY,
    ) -> File:
        repo = (
            repository
            if isinstance(repository, FileRepository)
            else get_builtin_repository(repository)
        )

        fdata = FileData(data, content_type, file_name)

        return cls(
            url=repo.save(fdata),
            content_type=fdata.content_type,
            file_name=fdata.file_name,
            file_size=len(data),
            file_data=data,
        )

    @classmethod
    def from_path(
        cls,
        path: str | Path,
        content_type: Optional[str] = None,
        repository: FileRepository | RepositoryId = DEFAULT_REPOSITORY,
        multipart: bool | None = None,
    ) -> File:
        file_path = Path(path)
        if not file_path.exists():
            raise FileNotFoundError(f"File {file_path} does not exist")

        repo = (
            repository
            if isinstance(repository, FileRepository)
            else get_builtin_repository(repository)
        )

        content_type = content_type or "application/octet-stream"

        url, data = repo.save_file(
            file_path,
            content_type=content_type,
            multipart=multipart,
        )
        return cls(
            url=url,
            file_data=data.data if data else None,
            content_type=content_type,
            file_name=file_path.name,
            file_size=file_path.stat().st_size,
        )

    def as_bytes(self) -> bytes:
        if self.file_data is None:
            raise ValueError("File has not been downloaded")

        return self.file_data

    def save(self, path: str | Path, overwrite: bool = False) -> Path:
        file_path = Path(path).resolve()

        if file_path.exists() and not overwrite:
            raise FileExistsError(f"File {file_path} already exists")

        downloaded_path = download_file(self.url, target_dir=file_path.parent)
        downloaded_path.rename(file_path)

        return file_path


class CompressedFile(File):
    extract_dir: Optional[str] = Field(default=None, exclude=True, repr=False)

    def __iter__(self):
        if not self.extract_dir:
            self._extract_files()

        files = Path(self.extract_dir).iterdir()  # type: ignore
        return iter(files)

    def _extract_files(self):
        self.extract_dir = mkdtemp()

        with NamedTemporaryFile() as temp_file:
            file_path = temp_file.name
            self.save(file_path, overwrite=True)

            with ZipFile(file_path) as zip_file:
                zip_file.extractall(self.extract_dir)

    def glob(self, pattern: str):
        if not self.extract_dir:
            self._extract_files()

        return Path(self.extract_dir).glob(pattern)  # type: ignore

    def __del__(self):
        if self.extract_dir:
            shutil.rmtree(self.extract_dir)



from fal.cli.auth import _login, _logout, _whoami
from fal.cli.main import parse_args


def test_login():
    args = parse_args(["auth", "login"])
    assert args.func == _login


def test_logout():
    args = parse_args(["auth", "logout"])
    assert args.func == _logout


def test_whoami():
    args = parse_args(["auth", "whoami"])
    assert args.func == _whoami



from __future__ import annotations

from fal import cached, function


class DepCass:
    def dep1(self):
        return "1"


def dep2():
    return "2"


@cached
def dep3():
    return "3"


@function("virtualenv", keep_alive=0, machine_type="XS")
def mainified():
    return DepCass().dep1() + dep2() + dep3()



from __future__ import annotations

from functools import lru_cache
from typing import TYPE_CHECKING
from urllib.request import Request, urlopen

from .image import *  # noqa: F403

if TYPE_CHECKING:
    from PIL.Image import Image as PILImage


def filter_by(
    has_nsfw_concepts: list[bool],
    images: list[PILImage],
) -> list[PILImage]:
    from PIL import Image as PILImage

    return [
        (
            PILImage.new("RGB", (image.width, image.height), (0, 0, 0))
            if has_nsfw
            else image
        )
        for image, has_nsfw in zip(images, has_nsfw_concepts)
    ]


def preprocess_image(image_pil, convert_to_rgb=True, fix_orientation=True):
    from PIL import ImageOps, ImageSequence

    # For MPO (multi picture object) format images, we only need the first image
    images = []
    for image in ImageSequence.Iterator(image_pil):
        img = image

        if convert_to_rgb:
            img = img.convert("RGB")

        if fix_orientation:
            img = ImageOps.exif_transpose(img)

        images.append(img)

        break

    return images[0]


@lru_cache(maxsize=64)
def read_image_from_url(
    url: str, convert_to_rgb: bool = True, fix_orientation: bool = True
):
    from fastapi import HTTPException
    from PIL import Image

    TEMP_HEADERS = {
        "User-Agent": (
            "Mozilla/5.0 (Macintosh; Intel Mac OS X 10.8; rv:21.0) "
            "Gecko/20100101 Firefox/21.0"
        ),
    }

    try:
        request = Request(url, headers=TEMP_HEADERS)
        response = urlopen(request)
        image_pil = Image.open(response)
    except Exception:
        import traceback

        traceback.print_exc()
        raise HTTPException(422, f"Could not load image from url: {url}")

    return preprocess_image(image_pil, convert_to_rgb, fix_orientation)



from __future__ import annotations

import inspect
import json
import os
import queue
import re
import threading
import time
import typing
from contextlib import asynccontextmanager, contextmanager
from typing import Any, Callable, ClassVar, Literal, TypeVar

import httpx
from fastapi import FastAPI

import fal.api
from fal._serialization import include_modules_from
from fal.api import RouteSignature
from fal.exceptions import RequestCancelledException
from fal.logging import get_logger
from fal.toolkit.file.providers import fal as fal_provider_module

REALTIME_APP_REQUIREMENTS = ["websockets", "msgpack"]

EndpointT = TypeVar("EndpointT", bound=Callable[..., Any])
logger = get_logger(__name__)


async def _call_any_fn(fn, *args, **kwargs):
    if inspect.iscoroutinefunction(fn):
        return await fn(*args, **kwargs)
    else:
        return fn(*args, **kwargs)


def wrap_app(cls: type[App], **kwargs) -> fal.api.IsolatedFunction:
    include_modules_from(cls)

    def initialize_and_serve():
        app = cls()
        app.serve()

    metadata = {}
    try:
        app = cls(_allow_init=True)
        metadata["openapi"] = app.openapi()
    except Exception:
        logger.warning("Failed to build OpenAPI specification for %s", cls.__name__)
        realtime_app = False
    else:
        routes = app.collect_routes()
        realtime_app = any(route.is_websocket for route in routes)

    kind = cls.host_kwargs.pop("kind", "virtualenv")
    if kind == "container":
        cls.host_kwargs.pop("resolver", None)

    wrapper = fal.api.function(
        kind,
        requirements=cls.requirements,
        machine_type=cls.machine_type,
        **cls.host_kwargs,
        **kwargs,
        metadata=metadata,
        exposed_port=8080,
        serve=False,
    )
    fn = wrapper(initialize_and_serve)
    fn.options.add_requirements(fal.api.SERVE_REQUIREMENTS)
    if realtime_app:
        fn.options.add_requirements(REALTIME_APP_REQUIREMENTS)

    return fn


class EndpointClient:
    def __init__(self, url, endpoint, signature, timeout: int | None = None):
        self.url = url
        self.endpoint = endpoint
        self.signature = signature
        self.timeout = timeout

        annotations = endpoint.__annotations__ or {}
        self.return_type = annotations.get("return") or None

    def __call__(self, data):
        with httpx.Client() as client:
            resp = client.post(
                self.url + self.signature.path,
                json=data.dict() if hasattr(data, "dict") else dict(data),
                timeout=self.timeout,
            )
            resp.raise_for_status()
            resp_dict = resp.json()

        if not self.return_type:
            return resp_dict

        return self.return_type(**resp_dict)


class AppClient:
    def __init__(
        self,
        cls,
        url,
        timeout: int | None = None,
    ):
        self.url = url
        self.cls = cls

        for name, endpoint in inspect.getmembers(cls, inspect.isfunction):
            signature = getattr(endpoint, "route_signature", None)
            if signature is None:
                continue
            endpoint_client = EndpointClient(
                self.url,
                endpoint,
                signature,
                timeout=timeout,
            )
            setattr(self, name, endpoint_client)

    @classmethod
    @contextmanager
    def connect(cls, app_cls):
        app = wrap_app(app_cls)
        info = app.spawn()
        _shutdown_event = threading.Event()

        def _print_logs():
            while not _shutdown_event.is_set():
                try:
                    log = info.logs.get(timeout=0.1)
                except queue.Empty:
                    continue
                print(log)

        _log_printer = threading.Thread(target=_print_logs, daemon=True)
        _log_printer.start()

        try:
            with httpx.Client() as client:
                retries = 100
                for _ in range(retries):
                    resp = client.get(info.url + "/health", timeout=60)

                    if resp.is_success:
                        break
                    elif resp.status_code not in (500, 404):
                        resp.raise_for_status()
                    time.sleep(0.1)

            client = cls(app_cls, info.url)
            yield client
        finally:
            info.stream.cancel()
            _shutdown_event.set()
            _log_printer.join()

    def health(self):
        with httpx.Client() as client:
            resp = client.get(self.url + "/health")
            resp.raise_for_status()
            return resp.json()


PART_FINDER_RE = re.compile(r"[A-Z][a-z]*")


def _to_fal_app_name(name: str) -> str:
    # Convert MyGoodApp into my-good-app
    return "-".join(part.lower() for part in PART_FINDER_RE.findall(name))


class App(fal.api.BaseServable):
    requirements: ClassVar[list[str]] = []
    machine_type: ClassVar[str] = "S"
    host_kwargs: ClassVar[dict[str, Any]] = {
        "_scheduler": "nomad",
        "_scheduler_options": {
            "storage_region": "us-east",
        },
        "resolver": "uv",
        "keep_alive": 60,
    }
    app_name: ClassVar[str]
    app_auth: ClassVar[Literal["private", "public", "shared"]] = "private"

    def __init_subclass__(cls, **kwargs):
        app_name = kwargs.pop("name", None) or _to_fal_app_name(cls.__name__)
        parent_settings = getattr(cls, "host_kwargs", {})
        cls.host_kwargs = {**parent_settings, **kwargs}
        cls.app_name = getattr(cls, "app_name", app_name)

        if cls.__init__ is not App.__init__:
            raise ValueError(
                "App classes should not override __init__ directly. "
                "Use setup() instead."
            )

    def __init__(self, *, _allow_init: bool = False):
        if not _allow_init and not os.getenv("IS_ISOLATE_AGENT"):
            raise NotImplementedError(
                "Running apps through SDK is not implemented yet."
            )

    @classmethod
    def get_endpoints(cls) -> list[str]:
        return [
            signature.path
            for _, endpoint in inspect.getmembers(cls, inspect.isfunction)
            if (signature := getattr(endpoint, "route_signature", None))
        ]

    def collect_routes(self) -> dict[RouteSignature, Callable[..., Any]]:
        return {
            signature: endpoint
            for _, endpoint in inspect.getmembers(self, inspect.ismethod)
            if (signature := getattr(endpoint, "route_signature", None))
        }

    @asynccontextmanager
    async def lifespan(self, app: FastAPI):
        await _call_any_fn(self.setup)
        try:
            yield
        finally:
            await _call_any_fn(self.teardown)

    def health(self):
        return {}

    def setup(self):
        """Setup the application before serving."""

    def teardown(self):
        """Teardown the application after serving."""

    def _add_extra_middlewares(self, app: FastAPI):
        @app.middleware("http")
        async def provide_hints_headers(request, call_next):
            response = await call_next(request)
            try:
                response.headers["X-Fal-Runner-Hints"] = ",".join(self.provide_hints())
            except NotImplementedError:
                # This lets us differentiate between apps that don't provide hints
                # and apps that provide empty hints.
                pass
            except Exception:
                from fastapi.logger import logger

                logger.exception(
                    "Failed to provide hints for %s",
                    self.__class__.__name__,
                )
            return response

        @app.middleware("http")
        async def set_global_object_preference(request, call_next):
            response = await call_next(request)
            try:
                fal_provider_module.GLOBAL_LIFECYCLE_PREFERENCE = request.headers.get(
                    "X-Fal-Object-Lifecycle-Preference"
                )
            except Exception:
                from fastapi.logger import logger

                logger.exception(
                    "Failed set a global lifecycle preference %s",
                    self.__class__.__name__,
                )
            return response

        @app.exception_handler(RequestCancelledException)
        async def value_error_exception_handler(
            request, exc: RequestCancelledException
        ):
            from fastapi.responses import JSONResponse

            # A 499 status code is not an officially recognized HTTP status code,
            # but it is sometimes used by servers to indicate that a client has closed
            # the connection without receiving a response
            return JSONResponse({"detail": str(exc)}, 499)

    def _add_extra_routes(self, app: FastAPI):
        @app.get("/health")
        def health():
            return self.health()

    def provide_hints(self) -> list[str]:
        """Provide hints for routing the application."""
        raise NotImplementedError


def endpoint(
    path: str, *, is_websocket: bool = False
) -> Callable[[EndpointT], EndpointT]:
    """Designate the decorated function as an application endpoint."""

    def marker_fn(callable: EndpointT) -> EndpointT:
        if hasattr(callable, "route_signature"):
            raise ValueError(
                f"Can't set multiple routes for the same function: {callable.__name__}"
            )

        callable.route_signature = RouteSignature(path=path, is_websocket=is_websocket)  # type: ignore
        return callable

    return marker_fn


def _fal_websocket_template(
    func: EndpointT,
    route_signature: RouteSignature,
) -> EndpointT:
    # A template for fal's realtime websocket endpoints to basically
    # be a boilerplate for the user to fill in their inference function
    # and start using it.

    import asyncio
    from collections import deque
    from contextlib import suppress

    import msgpack
    from fastapi import WebSocket, WebSocketDisconnect

    async def mirror_input(queue: deque[Any], websocket: WebSocket) -> None:
        while True:
            try:
                raw_input = await asyncio.wait_for(
                    websocket.receive_bytes(),
                    timeout=route_signature.session_timeout,
                )
            except asyncio.TimeoutError:
                return

            input = msgpack.unpackb(raw_input, raw=False)
            if route_signature.input_modal:
                input = route_signature.input_modal(**input)

            queue.append(input)

    async def mirror_output(
        self,
        queue: deque[Any],
        websocket: WebSocket,
    ) -> None:
        loop = asyncio.get_event_loop()
        max_allowed_buffering = route_signature.buffering or 1
        outgoing_messages: asyncio.Queue[bytes] = asyncio.Queue(
            maxsize=max_allowed_buffering * 2  # x2 for outgoing timings
        )

        async def emit(message):
            if isinstance(message, bytes):
                await websocket.send_bytes(message)
            elif isinstance(message, str):
                await websocket.send_text(message)
            else:
                raise TypeError(f"Can't send message of type {type(message)}")

        async def background_emitter():
            while True:
                output = await outgoing_messages.get()
                await emit(output)
                outgoing_messages.task_done()

        emitter = asyncio.create_task(background_emitter())

        while True:
            if not queue:
                await asyncio.sleep(0.05)
                continue

            input = queue.popleft()
            if input is None or emitter.done():
                if not emitter.done():
                    await outgoing_messages.join()
                    emitter.cancel()

                with suppress(asyncio.CancelledError):
                    await emitter
                return None  # End of input

            batch = [input]
            while queue and len(batch) < route_signature.max_batch_size:
                next_input = queue.popleft()
                if hasattr(input, "can_batch") and not input.can_batch(
                    next_input, len(batch)
                ):
                    queue.appendleft(next_input)
                    break
                batch.append(next_input)

            t0 = loop.time()
            output = await loop.run_in_executor(None, func, self, *batch)  # type: ignore
            total_time = loop.time() - t0
            if not isinstance(output, dict):
                # Handle pydantic output modal
                if hasattr(output, "dict"):
                    output = output.dict()
                else:
                    raise TypeError(
                        "Expected a dict or pydantic model as output, got "
                        f"{type(output)}"
                    )

            messages = [
                msgpack.packb(output, use_bin_type=True),
            ]
            if route_signature.emit_timings:
                # We emit x-fal messages in JSON, no matter what the
                # input/output format is.
                timings = {
                    "type": "x-fal-message",
                    "action": "timings",
                    "timing": total_time,
                }
                messages.append(json.dumps(timings, separators=(",", ":")))

            for message in messages:
                try:
                    outgoing_messages.put_nowait(message)
                except asyncio.QueueFull:
                    await emit(message)

    async def websocket_template(self, websocket: WebSocket) -> None:
        import asyncio

        await websocket.accept()

        queue: deque[Any] = deque(maxlen=route_signature.buffering)
        input_task = asyncio.create_task(mirror_input(queue, websocket))
        input_task.add_done_callback(lambda _: queue.append(None))
        output_task = asyncio.create_task(mirror_output(self, queue, websocket))

        try:
            await asyncio.wait(
                {
                    input_task,
                    output_task,
                },
                return_when=asyncio.FIRST_COMPLETED,
            )
            if input_task.done():
                # User didn't send any input within the timeout
                # so we can just close the connection after the
                # processing of the last input is done.
                input_task.result()
                await asyncio.wait_for(
                    output_task, timeout=route_signature.session_timeout
                )
            else:
                assert output_task.done()

                # The execution of the inference function failed or exitted,
                # so just propagate the result.
                input_task.cancel()
                with suppress(asyncio.CancelledError):
                    await input_task

                output_task.result()
        except WebSocketDisconnect:
            input_task.cancel()
            output_task.cancel()
            with suppress(asyncio.CancelledError):
                await input_task

            with suppress(asyncio.CancelledError):
                await output_task
        except Exception as exc:
            import traceback

            traceback.print_exc()

            await websocket.send_json(
                {
                    "type": "x-fal-error",
                    "error": "INTERNAL_ERROR",
                    "reason": str(exc),
                }
            )
        else:
            await websocket.send_json(
                {
                    "type": "x-fal-error",
                    "error": "TIMEOUT",
                    "reason": "no inputs, reconnect when needed!",
                }
            )

        await websocket.close()

    # Seems like templating + stringified annotations don't play well,
    # so we have to set them manually.
    websocket_template.__annotations__ = {
        "websocket": WebSocket,
        "return": None,
    }
    websocket_template.route_signature = route_signature  # type: ignore
    websocket_template.original_func = func  # type: ignore
    return typing.cast(EndpointT, websocket_template)


_SENTINEL = object()


def realtime(
    path: str,
    *,
    buffering: int | None = None,
    session_timeout: float | None = None,
    input_modal: Any | None = _SENTINEL,
    max_batch_size: int = 1,
) -> Callable[[EndpointT], EndpointT]:
    """Designate the decorated function as a realtime application endpoint."""

    def marker_fn(original_func: EndpointT) -> EndpointT:
        nonlocal input_modal

        if hasattr(original_func, "route_signature"):
            raise ValueError(
                "Can't set multiple routes for the same function: "
                f"{original_func.__name__}"
            )

        if input_modal is _SENTINEL:
            type_hints = typing.get_type_hints(original_func)
            if len(type_hints) >= 1:
                input_modal = type_hints[list(type_hints.keys())[0]]
            else:
                input_modal = None

        route_signature = RouteSignature(
            path=path,
            is_websocket=True,
            input_modal=input_modal,
            buffering=buffering,
            session_timeout=session_timeout,
            max_batch_size=max_batch_size,
        )
        return _fal_websocket_template(
            original_func,
            route_signature,
        )

    return marker_fn



from __future__ import annotations

from dataclasses import dataclass

from ._base import AppException

# PyTorch error message for out of memory
_CUDA_OOM_MESSAGE = "CUDA error: out of memory"

# Special status code for CUDA out of memory errors
_CUDA_OOM_STATUS_CODE = 503


@dataclass
class CUDAOutOfMemoryException(AppException):
    """Exception raised when a CUDA operation runs out of memory."""

    message: str = _CUDA_OOM_MESSAGE
    status_code: int = _CUDA_OOM_STATUS_CODE


# based on https://github.com/Lightning-AI/pytorch-lightning/blob/37e04d075a5532c69b8ac7457795b4345cca30cc/src/lightning/pytorch/utilities/memory.py#L49
def _is_cuda_oom_exception(exception: BaseException) -> bool:
    return _is_cuda_out_of_memory(exception) or _is_cudnn_snafu(exception)


# based on https://github.com/BlackHC/toma/blob/master/toma/torch_cuda_memory.py
def _is_cuda_out_of_memory(exception: BaseException) -> bool:
    return (
        isinstance(exception, RuntimeError)
        and len(exception.args) == 1
        and "CUDA" in exception.args[0]
        and "out of memory" in exception.args[0]
    )


# based on https://github.com/BlackHC/toma/blob/master/toma/torch_cuda_memory.py
def _is_cudnn_snafu(exception: BaseException) -> bool:
    # For/because of https://github.com/pytorch/pytorch/issues/4107
    return (
        isinstance(exception, RuntimeError)
        and len(exception.args) == 1
        and "cuDNN error: CUDNN_STATUS_NOT_SUPPORTED." in exception.args[0]
    )



from __future__ import annotations

import json
import webbrowser
from argparse import ArgumentParser
from collections import Counter
from dataclasses import dataclass, field
from typing import Any, Iterator, Union, cast

import graphlib
import rich
from openapi_fal_rest.api.workflows import (
    create_workflow as publish_workflow,
)
from openapi_fal_rest.models.http_validation_error import HTTPValidationError
from pydantic import BaseModel
from rich.syntax import Syntax

import fal
from fal import flags
from fal.exceptions import FalServerlessException
from fal.rest_client import REST_CLIENT

JSONType = Union[dict[str, Any], list[Any], str, int, float, bool, None, "Leaf"]
SchemaType = dict[str, Any]

VARIABLE_PREFIX = "$"
INPUT_VARIABLE_NAME = "input"

# Will be 1.0 once the server is finalized and anything <1.0
# is going to be rejected.
WORKFLOW_EXPORT_VERSION = "0.1"


class WorkflowSyntaxError(FalServerlessException):
    pass


class MisconfiguredGraphError(WorkflowSyntaxError):
    pass


def parse_leaf(raw_leaf: str) -> Leaf:
    """Parses a leaf (which is in the form of $variable.field.field_2[index] etc.)
    into a tree of Leaf objects."""
    raw_parts = raw_leaf.split(".")
    reference, *raw_parts = raw_parts
    if not reference.startswith(VARIABLE_PREFIX):
        raise WorkflowSyntaxError(
            f"Invalid leaf: {raw_leaf} (must start with a reference)"
        )

    leaf: Leaf = ReferenceLeaf(reference.removeprefix(VARIABLE_PREFIX))
    for raw_part in raw_parts:
        if raw_part.isdigit():
            leaf = IndexLeaf(leaf, int(raw_part))
        elif raw_part.isidentifier():
            leaf = AttributeLeaf(leaf, raw_part)
        else:
            raise WorkflowSyntaxError(
                f"Invalid leaf: {raw_leaf} (unexpected {raw_part})"
            )

    return leaf


def export_workflow_json(data: JSONType) -> JSONType:
    if isinstance(data, dict):
        return {k: export_workflow_json(v) for k, v in data.items()}
    elif isinstance(data, list):
        return [export_workflow_json(v) for v in data]
    elif isinstance(data, Leaf):
        return repr(data)
    else:
        return data


def import_workflow_json(data: JSONType) -> JSONType:
    if isinstance(data, dict):
        return {k: import_workflow_json(v) for k, v in data.items()}
    elif isinstance(data, list):
        return [import_workflow_json(v) for v in data]
    elif isinstance(data, str) and data.startswith(VARIABLE_PREFIX):
        return parse_leaf(data)
    else:
        return data


def iter_leaves(data: JSONType) -> Iterator[JSONType]:
    if isinstance(data, dict):
        for value in data.values():
            yield from iter_leaves(value)
    elif isinstance(data, list):
        for item in data:
            yield from iter_leaves(item)
    else:
        yield data


def depends(data: JSONType) -> set[str]:
    return {
        leaf.referee.id  # type: ignore
        for leaf in iter_leaves(data)
        if isinstance(leaf, Leaf)
    }


@dataclass
class Context:
    vars: dict[str, JSONType]

    def hydrate(self, input: JSONType) -> JSONType:
        if isinstance(input, dict):
            return {k: self.hydrate(v) for k, v in input.items()}
        elif isinstance(input, list):
            return [self.hydrate(v) for v in input]
        elif isinstance(input, Leaf):
            return input.execute(self)
        else:
            return input


@dataclass
class Leaf:
    def execute(self, context: Context) -> JSONType:
        raise NotImplementedError

    def __getattr__(self, name: str) -> AttributeLeaf:
        return AttributeLeaf(self, name)

    def __getitem__(self, index: int) -> IndexLeaf:
        return IndexLeaf(self, index)

    @property
    def referee(self) -> ReferenceLeaf:
        raise NotImplementedError


@dataclass
class AttributeLeaf(Leaf):
    leaf: Leaf
    attribute: str

    def execute(self, context: Context) -> JSONType:
        output = self.leaf.execute(context)
        assert isinstance(output, dict), f"{self.leaf!r} is not a dict"
        return output[self.attribute]

    def __repr__(self) -> str:
        return f"{self.leaf!r}.{self.attribute}"

    @property
    def referee(self) -> ReferenceLeaf:
        return self.leaf.referee


@dataclass
class IndexLeaf(Leaf):
    leaf: Leaf
    index: int

    def execute(self, context: Context) -> JSONType:
        output = self.leaf.execute(context)
        assert isinstance(output, list), f"{self.leaf!r} is not an array"
        return output[self.index]

    def __repr__(self) -> str:
        return f"{self.leaf!r}.{self.index}"

    @property
    def referee(self) -> ReferenceLeaf:
        return self.leaf.referee


@dataclass
class ReferenceLeaf(Leaf):
    id: str

    def execute(self, context: Context) -> JSONType:
        try:
            return context.vars[self.id]
        except KeyError:
            raise MisconfiguredGraphError(f"Variable {self.id!r} is not defined")

    def __repr__(self) -> str:
        return VARIABLE_PREFIX + self.id

    @property
    def referee(self) -> ReferenceLeaf:
        return self


@dataclass
class Node:
    id: str
    depends: set[str]

    @classmethod
    def from_json(cls, data: dict[str, Any]) -> Node:
        type = data.pop("type")
        if type == "display":
            return Display.from_json(data)
        elif type == "run":
            return Run.from_json(data)
        else:
            raise WorkflowSyntaxError(f"Invalid node type: {type}")

    def to_json(self) -> dict[str, Any]:
        raise NotImplementedError

    def execute(self, context: Context) -> JSONType:
        raise NotImplementedError


@dataclass
class Display(Node):
    fields: list[Leaf]

    @classmethod
    def from_json(cls, data: dict[str, Any]) -> Display:
        return cls(
            id=data["id"],
            depends=set(data["depends"]),
            fields=import_workflow_json(data["fields"]),  # type: ignore
        )

    def to_json(self) -> dict[str, Any]:
        return {
            "type": "display",
            "id": self.id,
            "depends": list(self.depends),
            "fields": export_workflow_json(self.fields),
        }

    def execute(self, context: Context) -> JSONType:
        for url in context.hydrate(self.fields):  # type: ignore
            if flags.DONT_OPEN_LINKS:
                print("Link:", url)
            else:
                webbrowser.open(url)


@dataclass
class Run(Node):
    app: str
    input: JSONType

    @classmethod
    def from_json(cls, data: dict[str, Any]) -> Run:
        return cls(
            id=data["id"],
            depends=set(data["depends"]),
            app=data["app"],
            input=import_workflow_json(data["input"]),
        )

    def execute(self, context: Context) -> JSONType:
        input = context.hydrate(self.input)
        assert isinstance(input, dict)
        return cast(JSONType, fal.apps.run(self.app, input))

    def to_json(self) -> dict[str, Any]:
        return {
            "type": "run",
            "id": self.id,
            "app": self.app,
            "depends": list(self.depends),
            "input": export_workflow_json(self.input),  # type: ignore
        }


@dataclass
class Workflow:
    name: str
    input_schema: SchemaType
    output_schema: SchemaType
    nodes: dict[str, Node] = field(default_factory=dict)
    output: dict[str, Any] | None = None
    _app_counter: Counter = field(default_factory=Counter)

    @classmethod
    def from_json(cls, data: dict[str, Any]) -> Workflow:
        data = import_workflow_json(data)  # type: ignore
        return cls(
            name=data["name"],
            input_schema=data["schema"]["input"],
            output_schema=data["schema"]["output"],
            nodes={
                node_id: Node.from_json(node_data)
                for node_id, node_data in data["nodes"].items()
            },
            output=data["output"],
        )

    def __post_init__(self) -> None:
        for node in self.nodes.values():
            if isinstance(node, Run):
                self._app_counter[node.app] += 1

    def _generate_node_id(self, app: str) -> str:
        self._app_counter[app] += 1
        return f"{app.replace('/', '_').replace('-', '_')}_{self._app_counter[app]}"

    def run(self, app: str, input: JSONType) -> ReferenceLeaf:
        node_id = self._generate_node_id(app)
        node = self.nodes[node_id] = Run(
            id=node_id,
            depends=depends(input),
            app=app,
            input=input,
        )
        return ReferenceLeaf(node.id)

    def display(self, *fields: Leaf) -> None:
        node_id = self._generate_node_id("display")
        self.nodes[node_id] = Display(
            node_id,
            depends=depends(list(fields)),
            fields=list(fields),
        )

    def set_output(self, output: JSONType) -> None:
        self.output = output  # type: ignore

    def execute(self, input: JSONType) -> JSONType:
        if not self.output:
            raise WorkflowSyntaxError(
                "Can't execute the workflow before the output is set."
            )

        context = Context({INPUT_VARIABLE_NAME: input})

        sorter = graphlib.TopologicalSorter(
            graph={
                node.id: node.depends - {INPUT_VARIABLE_NAME}
                for node in self.nodes.values()
            }
        )
        for node_id in sorter.static_order():
            node = self.nodes[node_id]
            context.vars[node_id] = node.execute(context)

        return context.hydrate(self.output)

    @property
    def input(self) -> ReferenceLeaf:
        return ReferenceLeaf(INPUT_VARIABLE_NAME)

    def to_json(self) -> dict[str, JSONType]:
        if not self.output:
            raise WorkflowSyntaxError(
                "Can't serialize the workflow before the output is set."
            )

        return {
            "name": self.name,
            "schema": {
                "input": self.input_schema,
                "output": self.output_schema,
            },
            "nodes": {node.id: node.to_json() for node in self.nodes.values()},
            "output": export_workflow_json(self.output),
            "version": WORKFLOW_EXPORT_VERSION,
        }

    to_dict = to_json

    def publish(self, title: str, *, is_public: bool = True):
        workflow_contents = publish_workflow.TypedWorkflow(
            name=self.name,
            title=title,
            contents=self,  # type: ignore
            is_public=is_public,
        )
        published_workflow = publish_workflow.sync(
            client=REST_CLIENT,
            json_body=workflow_contents,
        )
        if isinstance(published_workflow, Exception):
            raise published_workflow
        if isinstance(published_workflow, HTTPValidationError):
            raise RuntimeError(published_workflow.detail)
        if not published_workflow:
            raise RuntimeError("Failed to publish the workflow")

        # NOTE: dropping the provider prefix from the user_id
        user_id_part = published_workflow.user_id.split("|")[-1]
        return f"{user_id_part}/{published_workflow.name}"


def create_workflow(
    name: str,
    input: type[BaseModel],
    output: type[BaseModel],
) -> Workflow:
    return Workflow(
        name=name,
        input_schema=input.schema(),
        output_schema=output.schema(),
    )


def main() -> None:
    import cli_nested_json

    parser = ArgumentParser()
    parser.add_argument("workflow_file", type=str)
    args, input_params = parser.parse_known_args()

    with open(args.workflow_file) as stream:
        workflow = Workflow.from_json(json.load(stream))

    payload = cli_nested_json.interpret_nested_json(
        [part.split("=") for part in input_params]
    )
    console = rich.get_console()
    console.print(
        f"🤧 Loaded {workflow.name!r} with {len(workflow.nodes)} nodes!",
        style="bold magenta",
    )

    context = Context({INPUT_VARIABLE_NAME: payload})

    sorter = graphlib.TopologicalSorter(
        graph={
            node.id: node.depends - {INPUT_VARIABLE_NAME}
            for node in workflow.nodes.values()
        }
    )
    with console.status("Starting the execution", spinner="bouncingBall") as status:
        for n, node_id in enumerate(sorter.static_order()):
            node = workflow.nodes[node_id]
            status.update(
                status=f"Executing {node_id!r} ({n}/{len(workflow.nodes)})",
                spinner="runner",
            )
            if isinstance(node, Run):
                input = context.hydrate(node.input)
                assert isinstance(input, dict)

                owner, _, app = node.app.partition("/")
                app, sep, path = app.partition("/")

                handle = fal.apps.submit(
                    f"{owner}/{app}",
                    path=f"{sep}{path}",
                    arguments=input,
                )
                log_count = 0
                for event in handle.iter_events(logs=True):
                    if isinstance(event, fal.apps.Queued):
                        status.update(
                            status=(
                                "Queued for "
                                f"{node_id!r} "
                                f"(position={event.position}) "
                                f"({n}/{len(workflow.nodes)})",
                            ),
                            spinner="dots",
                        )
                    elif isinstance(event, fal.apps.InProgress):
                        status.update(
                            status=f"Executing {node_id!r} ({n}/{len(workflow.nodes)})",
                            spinner="runner",
                        )
                        for log in event.logs[log_count:]:  # type: ignore
                            console.log(log["message"], style="dim")
                            log_count += 1

                handle_status = handle.status(logs=True)
                assert isinstance(handle_status, fal.apps.Completed)
                for log in handle_status.logs[log_count:]:  # type: ignore
                    console.log(log["message"], style="dim")

                context.vars[node_id] = handle.get()
            else:
                context.vars[node_id] = node.execute(context)

        console.print(
            "🎉 Execution complete!",
            style="bold green",
        )
        output = context.hydrate(workflow.output)
        console.print(Syntax(json.dumps(output, indent=2), "json"))


if __name__ == "__main__":
    main()



from __future__ import annotations

import os
from dataclasses import dataclass, field

import click

from fal.auth import auth0, local
from fal.console import console
from fal.console.icons import CHECK_ICON
from fal.exceptions.auth import UnauthenticatedException


def key_credentials() -> tuple[str, str] | None:
    # Ignore key credentials when the user forces auth by user.
    if os.environ.get("FAL_FORCE_AUTH_BY_USER") == "1":
        return None

    if "FAL_KEY" in os.environ:
        key = os.environ["FAL_KEY"]
        key_id, key_secret = key.split(":", 1)
        return (key_id, key_secret)
    elif "FAL_KEY_ID" in os.environ and "FAL_KEY_SECRET" in os.environ:
        return (os.environ["FAL_KEY_ID"], os.environ["FAL_KEY_SECRET"])
    else:
        return None


def login():
    token_data = auth0.login()
    with local.lock_token():
        local.save_token(token_data["refresh_token"])

    USER.invalidate()


def logout():
    refresh_token, _ = local.load_token()
    if refresh_token is None:
        raise click.ClickException(message="You're not logged in")
    auth0.revoke(refresh_token)
    with local.lock_token():
        local.delete_token()

    USER.invalidate()
    console.print(f"{CHECK_ICON} Logged out of [cyan bold]fal[/]. Bye!")


def _fetch_access_token() -> str:
    """
    Load the refresh token, request a new access_token (refreshing the refresh token)
    and return the access_token.
    """
    # We need to lock both read and write access because we could be reading a soon
    # invalid refresh_token
    with local.lock_token():
        refresh_token, access_token = local.load_token()

        if refresh_token is None:
            raise UnauthenticatedException()

        if access_token is not None:
            try:
                auth0.verify_access_token_expiration(access_token)
                return access_token
            except Exception:
                # access_token expired, will refresh
                pass

        try:
            token_data = auth0.refresh(refresh_token)

            # NOTE: Auth0 Refresh Token Rotation enabled
            # So the old refresh_token is no longer valid
            local.save_token(token_data["refresh_token"], token_data["access_token"])
        except:
            local.delete_token()
            raise

        return token_data["access_token"]


@dataclass
class UserAccess:
    _access_token: str | None = field(repr=False, default=None)
    _user_info: dict | None = field(repr=False, default=None)
    _exc: Exception | None = field(repr=False, default=None)

    def invalidate(self) -> None:
        self._access_token = None
        self._user_info = None
        self._exc = None

    @property
    def info(self) -> dict:
        if self._user_info is None:
            self._user_info = auth0.get_user_info(self.bearer_token)

        return self._user_info

    @property
    def access_token(self) -> str:
        if self._exc is not None:
            # We access this several times, so we want to raise the
            # original exception instead of the newer exceptions we
            # would get from the effects of the original exception.
            raise self._exc

        if self._access_token is None:
            try:
                self._access_token = _fetch_access_token()
            except Exception as e:
                self._exc = e
                raise

        return self._access_token

    @property
    def bearer_token(self) -> str:
        return "Bearer " + self.access_token


USER = UserAccess()
